#include <typon/typon.hpp>
#include <typon/logger.hpp>
#include <cstdio>


using namespace typon;

int fibo(int n) {
  if (n < 2) {
    return n;
  }
  return fibo(n - 1) + fibo(n - 2);
}

Task<void> hello(int id) {
  int result = fibo(20);
  LOG("            || hello(%2d), fibo(20) = %d", id, result);
  co_return;
}

Task<void> fail() {
  throw std::exception();
  co_return;
}

Join<void> fail_join() {
  co_await fork(fail());
}

Join<void> join_sync(int max) {
  for (int id = 0; id < max; id++) {
    co_await fork(hello(id));
  }
  try {
    co_await fork(fail_join());
    LOG("    ... fork(fail_join()) did not throw (stolen continuation)");
  } catch(std::exception &) {
    LOG("    ... fork(fail_join()) did throw, rethrowing");
    throw;
  }
  LOG("    ... sync()");
  try
  {
    co_await Sync();
    LOG("    ... ERROR sync() did not throw");
  } catch(std::exception &) {
    LOG("    ... OK sync() did throw, rethrowing");
    throw;
  }
}

Join<void> join_no_sync(int max) {
  for (int id = 0; id < max; id++) {
    co_await fork(hello(id));
  }
  try {
    co_await fork(fail_join());
    LOG("    ... fork(fail_join()) did not throw (stolen continuation)");
  } catch(std::exception &) {
    LOG("    ... fork(fail_join()) did throw, rethrowing");
    throw;
  }
  LOG("    ... join_no_sync(%d) is exiting normally", max);
}

Root root_sync(int max) {
  try {
    LOG(">>> join_sync(%d)", max);
    co_await join_sync(max);
    LOG("ERROR join_sync(%d) did not throw\n", max);
  } catch(std::exception &) {
    LOG("OK join_sync(%d) did throw\n", max);
  }
}

Root root_no_sync(int max) {
  try {
    LOG(">>> join_no_sync(%d)", max);
    co_await join_no_sync(max);
    LOG("ERROR join_no_sync(%d) did not throw\n", max);
  } catch(std::exception &) {
    LOG("OK join_no_sync(%d) did throw\n", max);
  }
}

int main() {
  root_sync(0).call();
  root_sync(20).call();
  root_no_sync(0).call();
  root_no_sync(20).call();
  puts("done");
  return 0;
}

